package org.xbib.interlibrary.catalog.oracle.aleph;

import java.io.Closeable;
import java.io.Flushable;
import java.io.IOException;
import java.io.OutputStream;
import java.io.Reader;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.UncheckedIOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.EnumSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.interlibrary.api.action.worker.AbstractWorkerPool;
import org.xbib.interlibrary.api.action.worker.DefaultWorkerPoolListener;
import org.xbib.interlibrary.api.action.worker.Worker;
import org.xbib.interlibrary.api.action.worker.WorkerPool;
import org.xbib.interlibrary.api.action.worker.WorkerRequest;
import org.xbib.interlibrary.catalog.builder.ConverterOutput;
import org.xbib.interlibrary.catalog.builder.StringWorkerRequest;
import org.xbib.interlibrary.catalog.entities.ErrorFields;
import org.xbib.interlibrary.catalog.entities.ISILTransformer;
import org.xbib.interlibrary.catalog.entities.SkipFields;
import org.xbib.interlibrary.catalog.entities.TransformFields;
import org.xbib.marc.Marc;
import org.xbib.marc.MarcRecordAdapter;
import org.xbib.marc.MarcXchangeConstants;
import org.xbib.marc.json.MarcJsonWriter;
import org.xbib.marc.transformer.field.MarcFieldTransformers;
import org.xbib.marc.transformer.value.MarcValueTransformers;
import org.xbib.marc.xml.MarcContentHandler;
import org.xbib.settings.Settings;
import org.xml.sax.InputSource;

public class AlephPublishingConsumerPool extends AbstractWorkerPool<String> implements WorkerPool<String> {

    private static final Logger logger = Logger.getLogger(AlephPublishingConsumerPool.class.getName());

    private final Settings settings;

    private final MarcContentHandler marcContentHandler;

    private final List<AlephConnectorWorker> workers;

    private final ConverterOutput converterOutput;

    private final AlephMetric metric;

    public AlephPublishingConsumerPool(Settings settings) throws IOException {
        super(settings.get("name", "unnamed"),
                settings.getAsInt("workers", 1), new DefaultWorkerPoolListener<>());
        this.settings = settings;
        this.converterOutput = new ConverterOutput();
        converterOutput.createOutputFromFileMap(settings);
        this.marcContentHandler = createMarcContentHandlerFrom(settings, converterOutput);
        this.workers = new ArrayList<>();
        this.metric = new AlephMetric();
    }

    private static MarcContentHandler createMarcContentHandlerFrom(Settings settings, ConverterOutput converterOutput)
            throws IOException {
        logger.log(Level.INFO, "file output map = " + converterOutput.getMap());
        ConverterOutput.Entry entry = converterOutput.getMap().get("aleph");
        if (entry == null) {
            logger.log(Level.WARNING, "no output is defined, writing to a test JSON file");
            return new MarcJsonWriter(Files.newOutputStream(Paths.get("build/aleph-test.json")),
                    settings.getAsInt("buffersize", 65536))
                    .setStyle(EnumSet.of(MarcJsonWriter.Style.LINES))
                    .setFormat("AlephPublishing")
                    .setType(MarcXchangeConstants.BIBLIOGRAPHIC_TYPE);
        }
        OutputStream outputStream = entry.getOut();
        return new MarcJsonWriter(outputStream,
                settings.getAsInt("buffersize", 65536))
                .setStyle(EnumSet.of(MarcJsonWriter.Style.LINES))
                .setFormat("AlephPublishing")
                .setType(MarcXchangeConstants.BIBLIOGRAPHIC_TYPE);
    }

    public void up() throws IOException {
        super.open();
        AlephSysProducer alephSysProducer = new AlephSysProducer(settings.getAsSettings("input.aleph"), getQueue());
        alephSysProducer.prepareProduce();
        metric.start();
        try {
            alephSysProducer.produce();
        } catch (InterruptedException interruptedException) {
            throw new IOException(interruptedException);
        }
    }

    @Override
    public AlephConnectorWorker newWorker(String name, int count) {
        AlephConnectorWorker worker = new AlephConnectorWorker(settings.getAsSettings("input.aleph"), this);
        workers.add(worker);
        return worker;
    }

    @Override
    public WorkerRequest<String> newWorkerRequest(Object input, String origin) {
        return input != null ? new StringWorkerRequest(input.toString(), origin) : null;
    }

    @Override
    public void afterExecute(Worker<String> worker, WorkerRequest<String> request) {
    }

    public void down() throws IOException {
        // submit poisons, stop thread pools
        logger.info("shutting down");
        super.close();
    }

    @Override
    public void close() {
        // flush documents and close workers
        int returncode = 0;
        for (AlephConnectorWorker worker : workers) {
            logger.log(Level.INFO, "closing worker " + worker);
            worker.close();
        }
        if (!getExceptions().isEmpty()) {
            logger.log(Level.WARNING, "there were worker errors");
            returncode = -1;
        } else {
            logger.info("workers completed without errors");
        }
        if (marcContentHandler instanceof Flushable) {
            try {
                logger.info("flushing content handler");
                ((Flushable) marcContentHandler).flush();
            } catch (IOException e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
                returncode = -1;
            }
        }
        try {
            if (marcContentHandler instanceof Closeable) {
                logger.info("closing content handler");
                ((Closeable) marcContentHandler).close();
            }
        } catch (IOException e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
            returncode = -1;
        }
        logger.log(Level.INFO, marcContentHandler.getRecordCounter() + " MARC records written");
        try {
            // after close, perform retention
            converterOutput.closeFileMap(returncode);
        } catch (IOException e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
        }
        metric.close();
    }

    @Override
    public Iterator<WorkerRequest<String>> iterator() {
        return getQueue().iterator();
    }

    public static class AlephConnectorWorker implements Worker<String>, AutoCloseable {

        private final AlephPublishingConsumerPool pool;

        private final MarcContentHandler contentHandler;

        private final MarcFieldTransformers marcFieldTransformers;

        private final MarcValueTransformers marcValueTransformers;

        private final Connection connection;

        private final String query;

        private final String set;

        private final LinkedList<String> docnumbers;

        private final int batchsize;

        public AlephConnectorWorker(Settings settings, AlephPublishingConsumerPool alephPublishingConsumerPool) {
            this.pool = alephPublishingConsumerPool;
            try {
                this.connection = DriverManager.getConnection(settings.get("uri"),
                        settings.get("user"), settings.get("pass"));
                DatabaseMetaData databaseMetaData = connection.getMetaData();
                logger.log(Level.INFO, "Database: product name = " + databaseMetaData.getDatabaseProductName() +
                        " product version = " + databaseMetaData.getDatabaseProductVersion() +
                        " major version = " + databaseMetaData.getDatabaseMajorVersion() +
                        " minor version = " +  databaseMetaData.getDatabaseMinorVersion() );
                logger.log(Level.INFO, "Driver: name = " + databaseMetaData.getDriverName() +
                        " version = " + databaseMetaData.getDriverVersion() +
                        " major version = " + databaseMetaData.getDriverMajorVersion() +
                        " minor version = " + databaseMetaData.getDriverMinorVersion() );
                this.contentHandler = new MarcContentHandler();
                contentHandler.addNamespace("http://www.ddb.de/professionell/mabxml/mabxml-1.xsd");
                contentHandler.setFormat("AlephXML");
                contentHandler.setType("Bibliographic");
                // the MARCRecordAdapter does the magic that multiple threads can write to a single archive file
                contentHandler.setMarcListener(new MarcRecordAdapter(alephPublishingConsumerPool.marcContentHandler, null));
                // set up field transformers
                this.marcFieldTransformers = new MarcFieldTransformers();
                ErrorFields errorFields = new ErrorFields();
                errorFields.install(settings, marcFieldTransformers);
                SkipFields skipFields = new SkipFields();
                skipFields.install(settings, marcFieldTransformers);
                TransformFields transformFields = new TransformFields();
                transformFields.install(settings, marcFieldTransformers, marcFieldTransformers, marcFieldTransformers);
                // set up field value transformers
                this.marcValueTransformers = new MarcValueTransformers();
                ISILTransformer isilTransformer = new ISILTransformer(settings);
                isilTransformer.createTransformer(settings, marcValueTransformers);
                this.query = settings.get("query");
                this.set = settings.get("set");
                this.docnumbers = new LinkedList<>();
                this.batchsize = settings.getAsInt("batchsize", 1);
            } catch (Exception e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
                throw new UncheckedIOException(new IOException(e));
            }
        }

        @Override
        public void execute(WorkerRequest<String> workerRequest) throws IOException {
            String request = workerRequest.getRequest();
            docnumbers.add(request);
            if (docnumbers.size() >= batchsize) {
                processInputSources();
                docnumbers.clear();
            }
        }

        @Override
        public void close() {
            try {
                if (docnumbers.size() > 0) {
                    processInputSources();
                }
                connection.close();
            } catch (IOException | SQLException e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
            }
        }

        private void processInputSources() throws IOException {
            try {
                for (InputSource inputSource : getInputSources()) {
                    Marc.builder()
                            .setMarcFieldTransformers(marcFieldTransformers)
                            .setMarcValueTransformers(marcValueTransformers)
                            .setContentHandler(contentHandler)
                            .build()
                            .xmlReader().parse(inputSource);
                }
            } catch (SQLException e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
                throw new IOException(e);
            } catch (IOException e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
                throw e;
            }
        }

        private List<InputSource> getInputSources() throws SQLException, IOException {
            List<InputSource> inputSources = new ArrayList<>(docnumbers.size());
            for (String docnumber : docnumbers) {
                try (PreparedStatement stmt = prepareQuery(docnumber);
                     ResultSet results = stmt.executeQuery()) {
                    while (results.next()) {
                        Reader reader = results.getCharacterStream(1);
                        String string = results.getString(2);
                        if (reader != null) {
                            StringWriter sw = new StringWriter();
                            reader.transferTo(sw);
                            String s  = sw.toString();
                            pool.metric.inc();
                            pool.metric.incBytes(s.length());
                            inputSources.add(new InputSource(new StringReader(s)));
                        } else if (string != null) {
                            pool.metric.inc();
                            pool.metric.incBytes(string.length());
                            inputSources.add(new InputSource(new StringReader(string)));
                        } else {
                            logger.log(Level.WARNING, "nothing found for set " + set + " and doc number " + docnumber);
                        }
                    }
                }
            }
            return inputSources;
        }

        private PreparedStatement prepareQuery(String docNumber) throws SQLException {
            PreparedStatement preparedStatement = connection.prepareStatement(query);
            preparedStatement.setString(1, set);
            preparedStatement.setString(2, docNumber);
            return preparedStatement;
        }
    }
}
