package org.xbib.interlibrary.catalog.oracle.aleph;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.UncheckedIOException;
import java.nio.charset.StandardCharsets;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Collectors;
import org.xbib.interlibrary.api.action.worker.AbstractWorkerPool;
import org.xbib.interlibrary.api.action.worker.DefaultWorkerPoolListener;
import org.xbib.interlibrary.api.action.worker.Worker;
import org.xbib.interlibrary.api.action.worker.WorkerPool;
import org.xbib.interlibrary.api.action.worker.WorkerRequest;
import org.xbib.interlibrary.catalog.builder.ConverterOutput;
import org.xbib.interlibrary.catalog.builder.StringWorkerRequest;
import org.xbib.settings.Settings;

public class AlephSysDumperWorkerPool extends AbstractWorkerPool<String> implements WorkerPool<String> {

    private static final Logger logger = Logger.getLogger(AlephSysDumperWorkerPool.class.getName());

    private final Settings settings;

    private final List<AlephSysDumperWorker> workers;

    private final ConverterOutput converterOutput;

    private final AtomicLong existCounter;

    private final AtomicLong notExistCounter;

    public AlephSysDumperWorkerPool(Settings settings) throws IOException {
        super(settings.get("name", "unnamed"), getWorkerCount(settings), new DefaultWorkerPoolListener<>());
        this.settings = settings;
        this.converterOutput = new ConverterOutput();
        converterOutput.createOutputFromFileMap(settings);
        this.workers = new ArrayList<>();
        this.existCounter = new AtomicLong();
        this.notExistCounter = new AtomicLong();
    }

    private static int getWorkerCount(Settings settings) {
        return settings.getAsInt("workers", 1);
    }

    public void up() throws IOException {
        super.open();
        logger.log(Level.INFO, "workers: " + getWorkerCount(settings));
        Settings inputAlephSettings = settings.getAsSettings("input.aleph");
        AlephSysProducer alephSysProducer = new AlephSysProducer(inputAlephSettings, getQueue());
        try {
            alephSysProducer.produce();
        } catch (InterruptedException interruptedException) {
            throw new IOException(interruptedException);
        }
    }

    @Override
    public AlephSysDumperWorker newWorker(String name, int count) {
        AlephSysDumperWorker worker = new AlephSysDumperWorker(settings.getAsSettings("input.aleph"), this);
        workers.add(worker);
        return worker;
    }

    @Override
    public WorkerRequest<String> newWorkerRequest(Object input, String origin) {
        return input != null ? new StringWorkerRequest(input.toString(), origin) : null;
    }

    @Override
    public void afterExecute(Worker<String> worker, WorkerRequest<String> request) {
    }

    public void down() throws IOException {
        // submit poisons, stop thread pools
        logger.info("shutting down");
        super.close();
    }

    @Override
    public void close() throws IOException {
        // flush documents and close workers
        int returncode = 0;
        for (AlephSysDumperWorker worker : workers) {
            logger.log(Level.INFO, "closing worker " + worker);
            worker.close();
        }
        if (!getExceptions().isEmpty()) {
            logger.log(Level.WARNING, "there were worker errors");
            returncode = -1;
        } else {
            logger.info("workers completed without errors");
        }
        try {
            // after close, perform retention
            converterOutput.closeFileMap(returncode);
        } catch (IOException e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
        }
        logger.log(Level.INFO, MessageFormat.format("exist counter = {0} not exist counter = {1}", existCounter.get(), notExistCounter.get()));
    }

    @Override
    public Iterator<WorkerRequest<String>> iterator() {
        return getQueue().iterator();
    }

    public static class AlephSysDumperWorker implements Worker<String>, AutoCloseable {

        private final AlephSysDumperWorkerPool alephSysDumperWorkerPool;

        private final Connection connection;

        private final String query;

        private BufferedWriter writer;

        public AlephSysDumperWorker(Settings settings, AlephSysDumperWorkerPool alephSysDumperWorkerPool) {
            this.alephSysDumperWorkerPool = alephSysDumperWorkerPool;
            try {
                this.connection = DriverManager.getConnection(settings.get("uri"),
                        settings.get("user"), settings.get("pass"));
                this.query = settings.get("query");
            } catch (Exception e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
                throw new UncheckedIOException(new IOException(e));
            }
            if (alephSysDumperWorkerPool.converterOutput.getMap().get("dumper") != null) {
                ConverterOutput.Entry fileOutput = alephSysDumperWorkerPool.converterOutput.getMap().get("dumper");
                this.writer = new BufferedWriter(new OutputStreamWriter(fileOutput.getOut(), StandardCharsets.UTF_8));
            }
        }

        @Override
        public void execute(WorkerRequest<String> workerRequest) throws IOException {
            String request = workerRequest.getRequest();
            try (PreparedStatement stmt = prepareQuery(request);
                 ResultSet results = stmt.executeQuery()) {
                if (results != null && results.next()) {
                    if (results.getBytes(1) != null) {
                        String content;
                        try (BufferedReader reader =
                                     new BufferedReader(new InputStreamReader(new ByteArrayInputStream(results.getBytes(1)),
                                             StandardCharsets.UTF_8))) {
                            content = reader.lines().collect(Collectors.joining("\n"));
                        }
                        if (content.length() > 0) {
                            writer.write(request);
                            writer.write(System.lineSeparator());
                            alephSysDumperWorkerPool.existCounter.incrementAndGet();
                        } else {
                            logger.log(Level.WARNING, "empty content for sys number " + request);
                            alephSysDumperWorkerPool.notExistCounter.incrementAndGet();
                        }
                    } else {
                        String content;
                        try (BufferedReader reader = new BufferedReader(results.getCharacterStream(2))) {
                            content = reader.lines().collect(Collectors.joining("\n"));
                        }
                        if (content.length() == 0) {
                            logger.log(Level.WARNING, "empty content for sys number " + request);
                            alephSysDumperWorkerPool.notExistCounter.incrementAndGet();
                        }
                    }
                } else {
                    logger.log(Level.WARNING, "sys number does not exist in table " + request);
                    alephSysDumperWorkerPool.notExistCounter.incrementAndGet();
                }
            } catch (SQLException e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
                throw new IOException(e);
            }
        }

        @Override
        public void close() {
            try {
                connection.close();
            } catch (SQLException e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
            }
        }

        private PreparedStatement prepareQuery(String sysnumber) throws SQLException {
            PreparedStatement stmt = connection.prepareStatement(query);
            stmt.setObject(1, sysnumber, 999);
            return stmt;
        }
    }
}
