package org.xbib.interlibrary.catalog.oracle.aleph;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.text.MessageFormat;
import java.text.ParseException;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.concurrent.BlockingQueue;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.interlibrary.api.action.worker.WorkerRequest;
import org.xbib.interlibrary.catalog.builder.StringWorkerRequest;
import org.xbib.settings.Settings;
import org.xbib.time.chronic.Chronic;
import org.xbib.time.chronic.Span;

/**
 * Producer for Aleph sys numbers.
 */
class AlephSysProducer {

    private static final Logger logger = Logger.getLogger(AlephSysProducer.class.getName());

    private final Settings settings;

    private final BlockingQueue<WorkerRequest<String>> queue;

    private Path path;

    private Long sysnumber;

    private Long from;

    private Long to;

    AlephSysProducer(Settings settings, BlockingQueue<WorkerRequest<String>> queue) {
        this.settings = settings;
        this.queue = queue;
    }

    public void prepareProduce() {
        try (Connection connection = DriverManager.getConnection(settings.get("uri"),
                settings.get("user"), settings.get("pass"))) {
            prepareProduce(settings, connection);
        } catch (Exception e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
        }
    }

    public void produce() throws IOException, InterruptedException {
        if (sysnumber != null) {
            queue.put(new StringWorkerRequest(Long.toString(sysnumber), "sysproducer"));
            return;
        }
        if (from != null && to != null) {
            long l = from;
            while (l++ < to) {
                queue.put(new StringWorkerRequest(Long.toString(l), "sysproducer"));
            }
            return;
        }
        if (path != null) {
            try (BufferedReader reader = Files.newBufferedReader(path)) {
                String line;
                while ((line = reader.readLine()) != null) {
                    queue.put(new StringWorkerRequest(line, "sysproducer"));
                }
            }
        }
    }

    private void prepareProduce(Settings settings, Connection connection)
            throws SQLException, IOException {
        Path path = Paths.get(settings.get("sysnumberfile"));
        String set = settings.get("set");
        logger.log(Level.INFO, "set = " + set);
        String day = settings.get("day");
        if (day != null) {
            try {
                Span daySpan = Chronic.parse(day);
                ZoneId zoneId = settings.get("timezone") != null ? ZoneId.of(settings.get("timezone")) : ZoneId.systemDefault();
                String dayPattern = DateTimeFormatter.ofPattern("yyyyMMdd").withZone(zoneId).format(daySpan.getBeginCalendar());
                logger.log(Level.INFO, () -> MessageFormat.format("creating update for day {0} to path {1}", dayPattern, path));
                String query = settings.get("sysnumberquery");
                if (query != null) {
                    // we do not us bind parameters because JDBC driver is unable to cast to CHAR so that an index is used.
                    // instead we must use a) a like pattern and b) an index hint
                    prepareProduce(connection, String.format(query, "'" + set + "'", "'" + dayPattern + "%'"), path);
                } else {
                    logger.log(Level.SEVERE, "sysnumberquery not defined in settings, unable to continue");
                }
            } catch (ParseException e) {
                throw new IOException(e);
            }
        } else if (settings.get("sysnumber") != null) {
            String sysnumberStr = settings.get("sysnumber");
            logger.log(Level.INFO, MessageFormat.format("creating single number {0} ", sysnumberStr));
            sysnumber = Long.valueOf(sysnumberStr);
        } else if (settings.get("max") != null) {
            this.from = settings.getAsLong("min", 0L);
            this.to = settings.getAsLong("max", 0L);
        } else {
            String query = settings.get("sysnumberquery");
            if (query != null) {
                prepareProduce(connection, String.format(query, "'" + set + "'"), path);
            } else {
                logger.log(Level.SEVERE, "sysnumberquery not defined in settings, unable to continue");
            }
        }
    }

    /**
     * Read timestamped sys numbers from Aleph publishing, close query, and exercise the queue with the number.
     * @param connection connection
     * @param query query for query
     * @param path the path for the sys number output
     * @throws SQLException if SQL fails
     */
    private void prepareProduce(Connection connection, String query, Path path)
            throws IOException, SQLException {
        if (path == null) {
            logger.log(Level.WARNING, "sorry, no path");
            return;
        }
        // writing all sys numbers to file is faster than having opened a cursor for hours on Oracle,
        // this will avoid "snapshot too old" errors
        logger.log(Level.INFO, MessageFormat.format("start of fetching sys number, writing to {0}", path));
        long count = 0;
        Files.deleteIfExists(path);
        try (connection;
             BufferedWriter writer = Files.newBufferedWriter(path);
             PreparedStatement statement = prepareStatement(connection, query);
             ResultSet resultSet = statement.executeQuery()) {
            resultSet.setFetchSize(1000);
            while (resultSet.next()) {
                String s = resultSet.getString(1);
                writer.write(s);
                writer.write("\n");
                count++;
            }
        }
        logger.log(Level.INFO, MessageFormat.format("fetched {0} sys numbers to path {1}", count, path));
        this.path = path;
    }

    /**
     *
     * @param connection the connection
     * @param query the query
     * @return prepared statement
     * @throws SQLException if statement fails
     */
    private static PreparedStatement prepareStatement(Connection connection,
                                                      String query) throws SQLException {
        PreparedStatement statement = connection.prepareStatement(query);
        // set a reasonably high value so cursor fetch will be faster than normal
        statement.setFetchSize(1000);
        return statement;
    }
}
